# -*- coding: UTF-8 -*-
require 'AlmanacDateTime'
require 'AlmanacMath'

#
#= 黄経、朔日等、太陽および月の運行に関する計算用クラスライブラリ
#
#Authors::   Yoshihiko Hara(GoroNeko)
#Date::      2010/08/14
#Version::   0.9.0
#Copyright:: Copyright (C) Yoshihiko Hara(GoroNeko), 2010. All rights reserved.
#License::   GNU GENERAL PUBLIC LICENSE Version 2
#

class AlmanacAngleSunAndMoon

  # 本クラスライブラリの、名前とバージョン
  NAME = "AlmanacAngleSunAndMoon"
  VERSION = "0.9.0"

  # 1日あたりの月齢の変動角度定数
  # ※ 月齢は約29.5日で1周(365度移動)する
  CONSTANT_OF_PERIODIC_LUNAR_AGE = 12.3

  # ユリウス日（ajd）から太陽の黄経（elos）を求める
  #
  # 【引数】
  #   1.ユリウス日（ajd:Float）
  # 【戻り値】
  #   1.太陽の黄経（elos(ecliptic longitude of sun):Float）
  # 【備考】
  #   以下のプログラムは、海上保安庁水路部の略算式（『新こよみ便利帳
  #   （暦計算研究会編　恒星社厚生閣）』掲載）を参考に作成した。
  #
  def ajd_to_elos(ajd)

    # Almanac用算術クラスの準備
    almanac_math = AlmanacMath.new

    # t = ユリウス世紀数
    #     （J2000.0（A.D.2000年1月1日の正午（12UT）のJD）からの経過時間を、
    #     　36525日単位で表したもの）
    t = (ajd - 2451545.0) / 36525.0

    # 太陽の黄経（elos）を求める計算式
    elos = 280.4659 + 36000.7695 * t
    elos = elos + 1.9147     * almanac_math.degree_cos( 35999.05 * t + 267.52)
    elos = elos - 0.0048 * t * almanac_math.degree_cos( 35999.05 * t + 267.52)
    elos = elos + 0.0200     * almanac_math.degree_cos( 71998.1  * t + 265.1 )
    elos = elos + 0.0020     * almanac_math.degree_cos( 32964.0  * t + 158.0 )
    elos = elos + 0.0018     * almanac_math.degree_cos(    19.0  * t + 159.0 )
    elos = elos + 0.0018     * almanac_math.degree_cos(445267.0  * t + 208.0 )
    elos = elos + 0.0015     * almanac_math.degree_cos( 45038.0  * t + 254.0 )
    elos = elos + 0.0013     * almanac_math.degree_cos( 22519.0  * t + 352.0 )
    elos = elos + 0.0007     * almanac_math.degree_cos( 65929.0  * t +  45.0 )
    elos = elos + 0.0007     * almanac_math.degree_cos(  3035.0  * t + 110.0 )
    elos = elos + 0.0007     * almanac_math.degree_cos(  9038.0  * t +  64.0 )
    elos = elos + 0.0006     * almanac_math.degree_cos( 33718.0  * t + 316.0 )
    elos = elos + 0.0005     * almanac_math.degree_cos(   115.0  * t + 118.0 )
    elos = elos + 0.0005     * almanac_math.degree_cos(  2281.0  * t + 221.0 )
    elos = elos + 0.0004     * almanac_math.degree_cos( 29930.0  * t +  48.0 )
    elos = elos + 0.0004     * almanac_math.degree_cos( 31557.0  * t + 161.0 )

    # 太陽の黄経を"0 <= n <360"度内に収めるための補正を行う。
    elos = almanac_math.put_between_from_0_to_360(elos)

    return elos
  end

  # ユリウス日（ajd）から月の黄経（elom）を求める
  #
  # 【引数】
  #   1.ユリウス日（ajd:Float）
  # 【戻り値】
  #   1.月の黄経（elom(ecliptic longitude of moon):Float）クラス
  # 【備考】
  #   以下のプログラムは、海上保安庁水路部の略算式（『新こよみ便利帳
  #   （暦計算研究会編　恒星社厚生閣）』掲載）を参考に作成した。
  #
  def ajd_to_elom(ajd)

    # Almanac用算術クラスの準備
    almanac_math = AlmanacMath.new

    # t = ユリウス世紀数
    #     （J2000.0（A.D.2000年1月1日の正午（12UT）のJD）からの経過時間を、
    #     　36525日単位で表したもの）
    t = (ajd - 2451545.0) / 36525.0

    # 月の黄経（elom）を求める計算式
    elom = 218.3162 + 481267.8809 * t
    elom = elom + 6.2888 * almanac_math.degree_cos( 477198.868 * t +  44.963)
    elom = elom + 1.274  * almanac_math.degree_cos( 413335.35  * t +  10.74 )
    elom = elom + 0.6583 * almanac_math.degree_cos( 890534.22  * t + 145.7  )
    elom = elom + 0.2136 * almanac_math.degree_cos( 954397.74  * t + 179.93 )
    elom = elom + 0.1851 * almanac_math.degree_cos(  35999.05  * t +  87.53 )
    elom = elom + 0.1144 * almanac_math.degree_cos( 966404.0   * t + 276.5  )
    elom = elom + 0.0588 * almanac_math.degree_cos(  63863.5   * t + 124.2  )
    elom = elom + 0.0571 * almanac_math.degree_cos( 377336.3   * t +  13.2  )
    elom = elom + 0.0533 * almanac_math.degree_cos(1367733.1   * t + 280.7  )
    elom = elom + 0.0458 * almanac_math.degree_cos( 854535.2   * t + 148.2  )
    elom = elom + 0.0409 * almanac_math.degree_cos( 441199.8   * t +  47.4  )
    elom = elom + 0.0347 * almanac_math.degree_cos( 445267.1   * t +  27.9  )
    elom = elom + 0.0304 * almanac_math.degree_cos( 513197.9   * t + 222.5  )
    elom = elom + 0.0154 * almanac_math.degree_cos(  75870.0   * t +  41.0  )
    elom = elom + 0.0125 * almanac_math.degree_cos(1443603.0   * t +  52.0  )
    elom = elom + 0.011  * almanac_math.degree_cos( 489205.0   * t + 142.0  )
    elom = elom + 0.0107 * almanac_math.degree_cos(1303870.0   * t + 246.0  )
    elom = elom + 0.01   * almanac_math.degree_cos(1431597.0   * t + 315.0  )
    elom = elom + 0.0085 * almanac_math.degree_cos( 826671.0   * t + 111.0  )
    elom = elom + 0.0079 * almanac_math.degree_cos( 449334.0   * t + 188.0  )
    elom = elom + 0.0068 * almanac_math.degree_cos( 926533.0   * t + 323.0  )
    elom = elom + 0.0052 * almanac_math.degree_cos(  31932.0   * t + 107.0  )
    elom = elom + 0.005  * almanac_math.degree_cos( 481266.0   * t + 205.0  )
    elom = elom + 0.004  * almanac_math.degree_cos(1331734.0   * t + 283.0  )
    elom = elom + 0.004  * almanac_math.degree_cos(1844932.0   * t +  56.0  )
    elom = elom + 0.004  * almanac_math.degree_cos(    133.0   * t +  29.0  )
    elom = elom + 0.0038 * almanac_math.degree_cos(1781068.0   * t +  21.0  )
    elom = elom + 0.0037 * almanac_math.degree_cos( 541062.0   * t + 259.0  )
    elom = elom + 0.0028 * almanac_math.degree_cos(   1934.0   * t + 145.0  )
    elom = elom + 0.0027 * almanac_math.degree_cos( 918399.0   * t + 182.0  )
    elom = elom + 0.0026 * almanac_math.degree_cos(1379739.0   * t +  17.0  )
    elom = elom + 0.0024 * almanac_math.degree_cos(  99863.0   * t + 122.0  )
    elom = elom + 0.0023 * almanac_math.degree_cos( 922466.0   * t + 163.0  )
    elom = elom + 0.0022 * almanac_math.degree_cos( 818536.0   * t + 151.0  )
    elom = elom + 0.0021 * almanac_math.degree_cos( 990397.0   * t + 357.0  )
    elom = elom + 0.0021 * almanac_math.degree_cos(  71998.0   * t +  85.0  )
    elom = elom + 0.0021 * almanac_math.degree_cos( 341337.0   * t +  16.0  )
    elom = elom + 0.0018 * almanac_math.degree_cos( 401329.0   * t + 274.0  )
    elom = elom + 0.0016 * almanac_math.degree_cos(1856938.0   * t + 152.0  )
    elom = elom + 0.0012 * almanac_math.degree_cos(1267871.0   * t + 249.0  )
    elom = elom + 0.0011 * almanac_math.degree_cos(1920802.0   * t + 186.0  )
    elom = elom + 0.0009 * almanac_math.degree_cos( 858602.0   * t + 129.0  )
    elom = elom + 0.0008 * almanac_math.degree_cos(1403732.0   * t +  98.0  )
    elom = elom + 0.0007 * almanac_math.degree_cos( 790672.0   * t + 114.0  )
    elom = elom + 0.0007 * almanac_math.degree_cos( 405201.0   * t +  50.0  )
    elom = elom + 0.0007 * almanac_math.degree_cos( 485333.0   * t + 186.0  )
    elom = elom + 0.0007 * almanac_math.degree_cos(  27864.0   * t + 127.0  )
    elom = elom + 0.0006 * almanac_math.degree_cos( 111869.0   * t +  38.0  )
    elom = elom + 0.0006 * almanac_math.degree_cos(2258267.0   * t + 156.0  )
    elom = elom + 0.0005 * almanac_math.degree_cos(1908795.0   * t +  90.0  )
    elom = elom + 0.0005 * almanac_math.degree_cos(1745069.0   * t +  24.0  )
    elom = elom + 0.0005 * almanac_math.degree_cos( 509131.0   * t + 242.0  )
    elom = elom + 0.0004 * almanac_math.degree_cos(  39871.0   * t + 223.0  )
    elom = elom + 0.0004 * almanac_math.degree_cos(  12006.0   * t + 187.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos( 958465.0   * t + 340.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos( 381404.0   * t + 354.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos( 349472.0   * t + 337.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos(1808933.0   * t +  58.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos( 549197.0   * t + 220.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos(   4067.0   * t +  70.0  )
    elom = elom + 0.0003 * almanac_math.degree_cos(2322131.0   * t + 191.0  )

    # 月の黄経を"0 <= n <360"度内に収めるための補正を行う。
    elom = almanac_math.put_between_from_0_to_360(elom)

    return elom
  end

  # 直前の朔日（新月・月齢0日）の日時情報を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.朔日の日時情報(saku_date_time：AlmanacDateTime)
  # 【備考】
  #   朔　 → 月の視黄経 - 太陽の視黄経 = 0 となった瞬間
  #   朔日 → 朔を含む日
  #
  def get_saku_immediately_before_target_day(target_day)

    # 対象日～（対象日 - 30）日のいずれかの日が朔日
    # ※ 月齢は0～29
    # その日の始まり（0時0分0秒）の"月の黄経 - 太陽の黄経"の値と
    # その日の終わり（23時59分59秒）の"月の黄経 - 太陽の黄経"の値を
    # 比べて、値の＋－がひっくり返っている日が、朔を含んでいる日
    saku_date_time = target_day
    day_of_first \
      = AlmanacDateTime.new(target_day.year,
                            target_day.month,
                            target_day.day,
                            0,
                            0,
                            0,
                            target_day.offset)
    day_of_end \
      = AlmanacDateTime.new(target_day.year,
                            target_day.month,
                            target_day.day,
                            23,
                            59,
                            59,
                            target_day.offset)
    temporary_day = 0
    i = 0
    while i < 30
      subtract_first_elos_from_first_elom \
        = ajd_to_elom(day_of_first.ajd) - ajd_to_elos(day_of_first.ajd)
      subtract_end_elos_from_first_elom \
        = ajd_to_elom(day_of_end.ajd) - ajd_to_elos(day_of_end.ajd)

      if subtract_first_elos_from_first_elom >= 0 && \
         subtract_first_elos_from_first_elom < CONSTANT_OF_PERIODIC_LUNAR_AGE && \
         subtract_end_elos_from_first_elom < 0 && \
         subtract_end_elos_from_first_elom >= -CONSTANT_OF_PERIODIC_LUNAR_AGE
        temporary_day = i
        break
      end
      if subtract_first_elos_from_first_elom < 0 && \
         subtract_first_elos_from_first_elom >= -CONSTANT_OF_PERIODIC_LUNAR_AGE && \
         subtract_end_elos_from_first_elom >= 0 && \
         subtract_end_elos_from_first_elom < CONSTANT_OF_PERIODIC_LUNAR_AGE
        temporary_day = i
        break
      end

      i += 1
      day_of_first -= 1
      day_of_end -= 1
    end
    saku_date_time -= i

    return saku_date_time
  end

  # 直前の朔日（新月・月齢0日）からの経過日数を返す
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.経過日数(lapsed_days：Integer)
  def get_day_of_passage_from_saku(target_day)

    # 仮の節の日の格納域
    temporary_saku_day \
      = get_saku_immediately_before_target_day(target_day)

    # 節の日からの経過日数を算出
    # （対象日のajdから節の日のajdを引いて求める）
    lapsed_days = (target_day.ajd - temporary_saku_day.ajd).to_i

    return lapsed_days

  end
end
