# -*- coding: UTF-8 -*-
require 'AlmanacDateTime'
require 'AlmanacAngleSunAndMoon'

#
#= 二十四節気(Solar term)算出用クラスライブラリ
#
#Authors::   Yoshihiko Hara(GoroNeko)
#Date::      2010/08/22
#Version::   0.9.0
#Copyright:: Copyright (C) Yoshihiko Hara(GoroNeko), 2010. All rights reserved.
#License::   GNU GENERAL PUBLIC LICENSE Version 2
#

class AlmanacSolarTerm

  # 本クラスライブラリの、名前とバージョン
  NAME = "AlmanacSolarTerm"
  VERSION = "0.9.0"

  # 二十四節気のそれぞれの節気を示す
  # 0(立春)～23(大寒)
  SOLAR_TERM_NUMBER \
    = Array.new([0,1,2,3,4,5,
                 6,7,8,9,10,11,
                 12,13,14,15,16,17,
                 18,19,20,21,22,23])
  # 二十四節気の漢字和名
  SOLAR_TERM_NAME_J \
    = Array.new(["立春", "雨水", "啓蟄", "春分", "清明", "穀雨",
                 "立夏", "小満", "芒種", "夏至", "小暑", "大暑",
                 "立秋", "処暑", "白露", "秋分", "寒露", "霜降",
                 "立冬", "小雪", "大雪", "冬至", "小寒", "大寒"])
  # 二十四節気の英名
  # ウィキペディア英語版の記事を参考にした
  # (http://en.wikipedia.org/wiki/Solar_term)
  SOLAR_TERM_NAME_E \
    = Array.new(["start of spring", "rain water", "awakening of insects", "vernal equinox", "clear and bright", "grain rain",
                 "start of summer", "grain full", "grain in ear", "summer solstice", "minor heat", "major heat",
                 "start of autumn", "limit of heat", "	white dew", "autumnal equinox", "	cold dew", "frost descent",
                 "start of winter", "minor snow", "major snow", "winter solstice", "minor cold", "major cold"])
  # それぞれの節気の太陽黄経
  SOLAR_TERM_DEGREE \
    = Array.new([315.0, 330.0, 345.0, 360.0, 15.0, 30.0,
                 45.0, 60.0, 75.0, 90.0, 105.0, 120.0,
                 135.0, 150.0, 165.0, 180.0, 195.0, 210.0,
                 225.0, 240.0, 255.0, 270.0, 285.0, 300.0])
  # それぞれの節気が通常存在する月
  SOLAR_TERM_MONTH \
    = Array.new([2, 2, 3, 3, 4, 4, 
                 5, 5, 6, 6, 7, 7,
                 8, 8, 9, 9, 10, 10, 
                 11, 11, 12, 12, 1, 1])
  # 節気である可能性のある最初の日
  # ex.立春は2/3～2/6のどれかの日。故に立春である可能性のある最初の日は3日
  SOLAR_TERM_DAY \
    = Array.new([3, 17, 4, 19, 3, 19, 
                 4, 19, 4, 20, 5, 21,
                 6, 21, 6, 21, 6, 22, 
                 6, 21, 5, 20, 4, 19])
  # 節季が節である(true)、節でなく中である(false)の区別を返す
  DISTINCTION_OF_SETSU_AND_CHUU \
    = Array.new([true, false, true, false, true, false,
                 true, false, true, false, true, false,
                 true, false, true, false, true, false,
                 true, false, true, false, true, false])

  # 指定西暦年(year:Integer)かつ指定時差(of：Rational)の
  # 指定節気(solar_term_number：Integer)の
  # 年月日(sekki_date_time:AlmanacDateTime)を返す
  #
  # 【引数】
  #   1.対象年(year)
  #   2.時差(of)
  #   3.求める対象の節気(solar_term_number)
  # 【戻り値】
  #   1.節季の年月日情報(sekki_date_time)
  # 【備考】
  #   各節気は、節気である可能性のある最初の日(SOLAR_TERM_DAY)から+n日
  #   までのいづれかの日である
  #
  def get_sekki_date_time(year, of, solar_term_number)
    
    # 節気の日の各パラメタを確保（確保項目は以下のとおり）
    # 1.節気の日の太陽の黄経（度）
    # 2.節気の日の年
    # 3.節気の日の月
    # 4.節気の日の日
    # 5.節気の日の時
    # 6.節気の日の分
    # 7.節気の日の秒
    # 8.節気を求める場所の時差
    sekki_degree = SOLAR_TERM_DEGREE[solar_term_number]
    sekki_year = year
    sekki_month = SOLAR_TERM_MONTH[solar_term_number]
    sekki_day = SOLAR_TERM_DAY[solar_term_number]
    sekki_hour = 0
    sekki_minute = 0
    sekki_second = 0
    sekki_of = of

    # 節気である可能性のある最初の日が節気であったと仮定する。
    true_sekki_day = sekki_day
    temporary_elos = 0.0

    while sekki_degree >= temporary_elos

      # 節気であると仮定した日+1日の0時0分0秒の
      # 太陽の黄経を求める
      temporary_elos = get_elos(sekki_year, \
                                sekki_month, \
                                (true_sekki_day + 1), \
                                sekki_hour, \
                                sekki_minute, \
                                sekki_second, sekki_of)


      # 春分(solar_term_number = 3)の場合は
      # 以下の補正を実施
      if 3 == solar_term_number
        # 春分の黄経=0度(=360度)より、求められてくる黄経は、345 < x < 15
        # 求められた黄経が15度以下だった場合は補正をかけておく。
        if temporary_elos < 15.0
          temporary_elos += 360.0
        end
      end

      # 求めた太陽の黄経が、節気の日の太陽の黄経以下なら、まだ節気ではない
      if sekki_degree >= temporary_elos
        # 節気であると仮定した日+1日が節気であったと仮定する。
        true_sekki_day += 1
      end
    end

    # 確定した節気日時を使い、AlmanacDateTimeクラスを作成する
    sekki_date_time \
      = AlmanacDateTime.new(sekki_year, \
                            sekki_month, \
                            true_sekki_day, \
                            sekki_hour, \
                            sekki_minute, \
                            sekki_second, sekki_of)

    return sekki_date_time
  end

  # 指定されたパラメタから、任意の日時の太陽の黄経を求める
  #
  # 【引数】
  #   1.対象年(year)
  #   2.対象月(month)
  #   3.対象日(day)
  #   4.対象時(hour)
  #   5.対象分(minute)
  #   6.対象秒(second)
  #   7.対象時差(of)
  # 【戻り値】
  #   1.太陽の黄経(elso)
  # 【備考】
  #   実際の太陽の黄経度の算出は、AlmanacAngleSunAndMoonクラスの
  #   ajd_to_elosメソッドが行う
  #
  def get_elos(year, month, day, hour, minute, second, of)

    # AlmanacDateTimeオブジェクトを、組み立てる
    sekki_date_time \
      = AlmanacDateTime.new(year, \
                            month, \
                            day, \
                            hour, \
                            minute, \
                            second, of)

    temporary_elos \
      = AlmanacAngleSunAndMoon.new.ajd_to_elos(sekki_date_time.ajd)

    return temporary_elos
  end

  # 指定日時の節である節気番号（SOLAR_TERM_NUMBER）を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.節番号(setsu_number：Integer)
  # 【備考】
  #   対象日が節の日でなかった場合は、節気番号として99（エラー）を返す
  #
  def get_setsu_number(target_day)

    # 月ごとの、節の候補
    # 1月の場合、節は"小寒"(SOLAR_TERM_NUMBER=22)
    # 2月の場合、節は"立春"(SOLAR_TERM_NUMBER=0)
    # 3月の場合、節は"啓蟄"(SOLAR_TERM_NUMBER=2)
    # 4月の場合、節は"清明"(SOLAR_TERM_NUMBER=4)
    # 5月の場合、節は"立夏"(SOLAR_TERM_NUMBER=6)
    # 6月の場合、節は"芒種"(SOLAR_TERM_NUMBER=8)
    # 7月の場合、節は"小暑"(SOLAR_TERM_NUMBER=10)
    # 8月の場合、節は"立秋"(SOLAR_TERM_NUMBER=12)
    # 9月の場合、節は"白露"(SOLAR_TERM_NUMBER=14)
    # 10月の場合、節は"寒露"(SOLAR_TERM_NUMBER=16)
    # 11月の場合、節は"立冬"(SOLAR_TERM_NUMBER=18)
    # 12月の場合、節は"大雪"(SOLAR_TERM_NUMBER=20)
    setsu_number = 2 * (target_day.month - 2)
    if 0 > setsu_number
      setsu_number = 22
    end

    # 対象日と同年の上記で求めた節気番号の節の日が一致していることを
    # 確認する。
    setsu_day = get_sekki_date_time(target_day.year, \
                                    target_day.offset, \
                                    setsu_number)
    if setsu_day != target_day
      setsu_number = 99
    end

    return setsu_number
  end

  # 直前の節の日時情報を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.節の日時情報(setsu_date_time：AlmanacDateTime)
  #
  def get_setsu_immediately_before_target_day(target_day)

    # 節の日の格納域
    setsu_date_time = AlmanacDateTime.now

    # 月ごとの、直前の節候補
    # 1月の場合、直前の節は"小寒"(SOLAR_TERM_NUMBER=22)
    # または"大雪"(SOLAR_TERM_NUMBER=20)
    # 2月の場合、直前の節は"立春"(SOLAR_TERM_NUMBER=0)
    # または"小寒"(SOLAR_TERM_NUMBER=22)
    # 3月の場合、直前の節は"啓蟄"(SOLAR_TERM_NUMBER=2)
    # または"立春"(SOLAR_TERM_NUMBER=0)
    # 4月の場合、直前の節は"清明"(SOLAR_TERM_NUMBER=4)
    # または"啓蟄"(SOLAR_TERM_NUMBER=2)
    # 5月の場合、直前の節は"立夏"(SOLAR_TERM_NUMBER=6)
    # または"清明"(SOLAR_TERM_NUMBER=4)
    # 6月の場合、直前の節は"芒種"(SOLAR_TERM_NUMBER=8)
    # または"立夏"(SOLAR_TERM_NUMBER=6)
    # 7月の場合、直前の節は"小暑"(SOLAR_TERM_NUMBER=10)
    # または"芒種"(SOLAR_TERM_NUMBER=8)
    # 8月の場合、直前の節は"立秋"(SOLAR_TERM_NUMBER=12)
    # または"小暑"(SOLAR_TERM_NUMBER=10)
    # 9月の場合、直前の節は"白露"(SOLAR_TERM_NUMBER=14)
    # または"立秋"(SOLAR_TERM_NUMBER=12)
    # 10月の場合、直前の節は"寒露"(SOLAR_TERM_NUMBER=16)
    # または"白露"(SOLAR_TERM_NUMBER=14)
    # 11月の場合、直前の節は"立冬"(SOLAR_TERM_NUMBER=18)
    # または"寒露"(SOLAR_TERM_NUMBER=16)
    # 12月の場合、直前の節は"大雪"(SOLAR_TERM_NUMBER=20)
    # または"立冬"(SOLAR_TERM_NUMBER=18)
    target_solar_term_number = 2 * (target_day.month - 2)
    if 0 > target_solar_term_number
      target_solar_term_number = 22
    end
    other_target_solar_term_number  = 2 * (target_day.month - 3)
    if 0 > other_target_solar_term_number
      other_target_solar_term_number  = 20 + 2 * (target_day.month - 1)
    end

    # 直前の節の日を求める
    setsu_date_time \
      = get_sekki_date_time( \
          target_day.year, \
          target_day.offset, \
          target_solar_term_number)
    if setsu_date_time.day > target_day.day
      if other_target_solar_term_number == 20
        target_day.year -= 1
      end
      setsu_date_time \
        = get_sekki_date_time( \
            target_day.year, \
            target_day.offset, \
            other_target_solar_term_number)
    end

    return setsu_date_time
  end

  # 直前の節からの経過日数を返す
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.経過日数(lapsed_days：Integer)
  #
  def get_day_of_passage_from_setsu(target_day)

    # 仮の節の日の格納域
    temporary_setsu_day \
      = get_setsu_immediately_before_target_day(target_day)

    # 節の日からの経過日数を算出
    # （対象日のajdから節の日のajdを引いて求める）
    lapsed_days = (target_day.ajd - temporary_setsu_day.ajd).to_i

    return lapsed_days

  end

  # 指定日時の中である節気番号（SOLAR_TERM_NUMBER）を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.中番号(cyuu_number：Integer)
  # 【備考】
  #   対象日が中の日でなかった場合は、節気番号として99（エラー）を返す
  #
  def get_cyuu_number(target_day)

    # 月ごとの、中の候補
    # 1月の場合、中は"大寒"(SOLAR_TERM_NUMBER=23)
    # 2月の場合、中は"雨水"(SOLAR_TERM_NUMBER=1)
    # 3月の場合、中は"春分"(SOLAR_TERM_NUMBER=3)
    # 4月の場合、中は"穀雨"(SOLAR_TERM_NUMBER=5)
    # 5月の場合、中は"小満"(SOLAR_TERM_NUMBER=7)
    # 6月の場合、中は"夏至"(SOLAR_TERM_NUMBER=9)
    # 7月の場合、中は"大暑"(SOLAR_TERM_NUMBER=11)
    # 8月の場合、中は"処暑"(SOLAR_TERM_NUMBER=13)
    # 9月の場合、中は"秋分"(SOLAR_TERM_NUMBER=15)
    # 10月の場合、中は"霜降"(SOLAR_TERM_NUMBER=17)
    # 11月の場合、中は"小雪"(SOLAR_TERM_NUMBER=19)
    # 12月の場合、中は"冬至"(SOLAR_TERM_NUMBER=21)
    cyuu_number = 1 + 2 * (target_day.month - 2)
    if 0 > cyuu_number
      cyuu_number = 23
    end

    # 対象日と同年の上記で求めた中番号の中の日が一致していることを
    # 確認する。
    cyuu_day = get_sekki_date_time(target_day.year, \
                                   target_day.offset, \
                                   cyuu_number)
    if cyuu_day != target_day
      cyuu_number = 99
    end

    return cyuu_number
  end

  # 直前の中の日時情報を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.中の日時情報(cyuu_date_time：AlmanacDateTime)
  #
  def get_cyuu_immediately_before_target_day(target_day)

    # 中の日の格納域
    cyuu_date_time = AlmanacDateTime.now

    # 月ごとの、直前の中候補
    # 1月の場合、直前の中は"大寒"(SOLAR_TERM_NUMBER=23)
    # または"冬至"(SOLAR_TERM_NUMBER=21)
    # 2月の場合、直前の中は"雨水"(SOLAR_TERM_NUMBER=1)
    # または"大寒"(SOLAR_TERM_NUMBER=23)
    # 3月の場合、直前の中は"春分"(SOLAR_TERM_NUMBER=3)
    # または"雨水"(SOLAR_TERM_NUMBER=1)
    # 4月の場合、直前の中は"穀雨"(SOLAR_TERM_NUMBER=5)
    # または"春分"(SOLAR_TERM_NUMBER=3)
    # 5月の場合、直前の中は"小満"(SOLAR_TERM_NUMBER=7)
    # または"穀雨"(SOLAR_TERM_NUMBER=5)
    # 6月の場合、直前の中は"夏至"(SOLAR_TERM_NUMBER=9)
    # または"小満"(SOLAR_TERM_NUMBER=7)
    # 7月の場合、直前の中は"大暑"(SOLAR_TERM_NUMBER=11)
    # または"夏至"(SOLAR_TERM_NUMBER=9)
    # 8月の場合、直前の中は"処暑"(SOLAR_TERM_NUMBER=13)
    # または"大暑"(SOLAR_TERM_NUMBER=11)
    # 9月の場合、直前の中は"秋分"(SOLAR_TERM_NUMBER=15)
    # または"処暑"(SOLAR_TERM_NUMBER=13)
    # 10月の場合、直前の中は"霜降"(SOLAR_TERM_NUMBER=17)
    # または"秋分"(SOLAR_TERM_NUMBER=15)
    # 11月の場合、直前の中は"小雪"(SOLAR_TERM_NUMBER=19)
    # または"霜降"(SOLAR_TERM_NUMBER=17)
    # 12月の場合、直前の中は"冬至"(SOLAR_TERM_NUMBER=21)
    # または"小雪"(SOLAR_TERM_NUMBER=19)
    target_solar_term_number = 1 + 2 * (target_day.month - 2)
    if 0 > target_solar_term_number
      target_solar_term_number = 23
    end
    other_target_solar_term_number  = 1 + 2 * (target_day.month - 3)
    if 0 > other_target_solar_term_number
      other_target_solar_term_number  = 21 + 2 * (target_day.month - 1)
    end

    # 直前の中の日を求める
    cyuu_date_time \
      = get_sekki_date_time( \
          target_day.year, \
          target_day.offset, \
          target_solar_term_number)
    if cyuu_date_time.day > target_day.day
      if other_target_solar_term_number == 21
        target_day.year -= 1
      end
      cyuu_date_time \
        = get_sekki_date_time( \
            target_day.year, \
            target_day.offset, \
            other_target_solar_term_number)
    end

    return cyuu_date_time
  end

  # 直前の中からの経過日数を返す
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.経過日数(lapsed_days：Integer)
  #
  def get_day_of_passage_from_cyuu(target_day)

    # 仮の中の日の格納域
    temporary_cyuu_day \
      = get_cyuu_immediately_before_target_day(target_day)

    # 中の日からの経過日数を算出
    # （対象日のajdから節の日のajdを引いて求める）
    lapsed_days = (target_day.ajd - temporary_cyuu_day.ajd).to_i

    return lapsed_days

  end

  # 指定日時の二至二分番号（SOLAR_TERM_NUMBER）を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.二至二分番号(nishinibun_number：Integer)
  # 【備考】
  #   二至二分とは、二至（夏至・冬至）および二分（春分・秋分）を指す
  #   対象日が二至二分の日でなかった場合は、中番号として99（エラー）を返す
  #
  def get_nishinibun_number(target_day)

    # 対象となる二至二分の候補
    # 3月の場合、二至二分は"春分"(SOLAR_TERM_NUMBER=3)
    # 6月の場合、二至二分は"夏至"(SOLAR_TERM_NUMBER=9)
    # 9月の場合、二至二分は"秋分"(SOLAR_TERM_NUMBER=15)
    # 12月の場合、二至二分は"冬至"(SOLAR_TERM_NUMBER=21)
    nishinibun_number = target_day.month + (target_day.month - 3)

    # 対象日と同年の上記で求めた二至二分節番号の
    # 二至二分の日が一致していることを
    # 確認する。
    nishinibun_day = get_sekki_date_time(target_day.year, \
                                         target_day.offset, \
                                         nishinibun_number)

    if nishinibun_day != target_day
      nishinibun_number = 99
    end

    return nishinibun_number
  end

  # 直前の二至二分の日時情報を求める
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.二至二分の日時情報(nishinibun_date_time：AlmanacDateTime)
  #
  def get_nishinibun_immediately_before_target_day(target_day)

    # 二至二分の日の格納域
    nishinibun_date_time = AlmanacDateTime.now

    # 直前の二至二分となりうる二至二分の候補は以下の5件
    # 以下の5件と指定日のajdを比較し、直前の二至二分を探し出す
    # 前年12月の"冬至"(SOLAR_TERM_NUMBER=21)
    # 今年3月の"春分"(SOLAR_TERM_NUMBER=3)
    # 今年6月の"夏至"(SOLAR_TERM_NUMBER=9)
    # 今年9月の"秋分"(SOLAR_TERM_NUMBER=15)
    # 今年12月の"冬至"(SOLAR_TERM_NUMBER=21)

    last_year_touji_date_time \
      = get_sekki_date_time((target_day.year - 1), target_day.offset, 21)
    current_year_shunbun_date_time \
      = get_sekki_date_time(target_day.year, target_day.offset, 3)
    current_year_geshi_date_time \
      = get_sekki_date_time(target_day.year, target_day.offset, 9)
    current_year_shuubun_date_time \
      = get_sekki_date_time(target_day.year, target_day.offset, 15)
    current_year_touji_date_time \
      = get_sekki_date_time(target_day.year, target_day.offset, 21)

    temporary_date_time_array = Array.new([current_year_touji_date_time, \
                                           current_year_shuubun_date_time, \
                                           current_year_geshi_date_time, \
                                           current_year_shunbun_date_time, \
                                           last_year_touji_date_time])

    i = 4
    nishinibun_date_time = temporary_date_time_array[4]
    while target_day.ajd > temporary_date_time_array[i].ajd
      nishinibun_date_time = temporary_date_time_array[i]
      i -= 1
    end

    return nishinibun_date_time
  end

  # 直前の二至二分からの経過日数を返す
  #
  # 【引数】
  #   1.対象日(target_day：AlmanacDateTime)
  # 【戻り値】
  #   1.経過日数(lapsed_days：Integer)
  #
  def get_day_of_passage_from_nishinibun(target_day)

    # 仮の二至二分の日の格納域
    temporary_nishinibun_day \
      = get_nishinibun_immediately_before_target_day(target_day)

    # 節の日からの経過日数を算出
    # （対象日のajdから節の日のajdを引いて求める）
    lapsed_days = (target_day.ajd - temporary_nishinibun_day.ajd).to_i

    return lapsed_days

  end
end
