# -*- coding: UTF-8 -*-
require "date"
require "curses"
require 'AlmanacDateTime'
require "AlmanacTwentyEightMansions"

#= 干支表示付き万年カレンダ スクリプト
#== AlmanacOrientalZodiac class の使用例
#
#Authors::   Yoshihiko Hara(GoroNeko)
#Date::      2010/08/23
#Version::   0.9.0
#Copyright:: Copyright (C) Yoshihiko Hara(GoroNeko), 2010. All rights reserved.
#License::   GNU GENERAL PUBLIC LICENSE Version 2
#

class TwentyEightMansionsCalendar

  # 本スクリプトの、名前とバージョン
  NAME = "OrientalZodiacCalendar"
  VERSION = "0.9.0"

  # カレンダー雛形生成用テーブル
  WEEK_TABLE = [
    [99, 99, 99, 99, 99, 99,  1,  2,  3,  4,  5,  6,  7],
    [ 2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14],
    [ 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21],
    [16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28],
    [23, 24, 25, 26, 27, 28, 29, 30, 31, 99, 99, 99, 99],
    [30, 31, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99],
  ]

  # 指定西暦年(year:Integer)・指定月(month：Integer)の
  # 1ヶ月分の二十八宿情報付きカレンダーを、コンソールに表示する
  #
  # 【引数】
  #   1.対象年(year)
  #   2.対象月(month)
  # 【戻り値】
  #   なし
  # 【備考】
  #   処理結果（カレンダの作成）を、直接コンソールに表示する
  #   コンソールへの表示には、Curses ライブラリを使用する
  #
  def print_calendar(year, month)

    # 指定された月の初日の情報をもとに
    # 最終日が何日かと、
    # カレンダ上の初日の表示位置とを求める
    first = Date.new(year, month, 1)
    end_of_month = ((first >> 1) - 1).day
    start = 6 - first.wday

    # 二十八宿 情報を算出する準備を行う
    work_date = AlmanacDateTime.new(year,month,1,12,0,0,Rational(3,8))
    twentyeight_mansions = AlmanacTwentyEightMansions.new

    # コンソールの初期化
    # ※ Curses ライブラリを使用
    Curses::init_screen

    # 年月日情報の編集
    title_string = "             " + year.to_s + "年 " + month.to_s + "月" + "\n"

    # 年月日情報を表示バッファにセット
    Curses::addstr(title_string)

    # 曜日情報の表示バッファにセット
    Curses::addstr(" 日   月   火   水   木   金   土  " + "\n")

    # 日付 および 二十八宿 情報をカレンダ形式に編集し
    # 表示バッファにセット
    WEEK_TABLE.each{|week|
      week_buf = ""
      twentyeight_mansions_buf = ""
      week[start, 7].each{|day|
        if day > end_of_month
          week_buf << "     "
          twentyeight_mansions_buf << "     "
        else
          work_week = sprintf("%3d", day) + "  "
          week_buf << work_week
          work_twentyeight_mansions_string \
            = AlmanacTwentyEightMansions::TWENTY_EIGHT_MANSIONS_NAME_J[
                twentyeight_mansions.get_twenty_eight_mansions_number(work_date)] + "宿 "
          twentyeight_mansions_buf << work_twentyeight_mansions_string
          work_date = work_date + 1
        end
      }
      week_buf = week_buf +"\n"
      Curses::addstr(week_buf)
      twentyeight_mansions_buf = twentyeight_mansions_buf + "\n"
      Curses::addstr(twentyeight_mansions_buf)
    }

    # コンソール画面に表示を実施
    Curses::refresh
    Curses::getch
    Curses::close_screen
  end
end

#
# --- << main >> ---
#

target_year = 99
target_month = 99


# ヘルプの表示を指定されていた場合は
# usage の表示のみを行い、処理を終了する
first_arg = ARGV[0].to_s
if "-?" == first_arg or
   "-h" == first_arg or
   "--help" == first_arg
  p "[usage]"
  p "> ruby TwentyEightMansionsCalendar.rb [year( 0 < yyyy < 10000 )] [month( 0 < mm < 13 )]"
  exit
end

# 処理に対し有効なパラメタ
# （ 第一引数 西暦 1 年 ～ 9999 年、
#    第二引数 1 月 ～ 12 月 ）が指定されていた場合、
# 指定年月のカレンダを作成し、コンソールに表示する
begin
  target_year = ARGV[0].to_i
  target_month = ARGV[1].to_i
rescue
  work_date = Date.today
  target_year = work_date.year
  target_month = work_date.month
end

if 9999 < target_year or target_year < 1 or
   12 < target_month or target_month < 1
  work_date = Date.today
  target_year = work_date.year
  target_month = work_date.month
end

targe_calendar = TwentyEightMansionsCalendar.new
targe_calendar.print_calendar(target_year, target_month)
