# -*- coding: UTF-8 -*-
require 'AlmanacDateTime'
require "AlmanacOrientalZodiac"
require "AlmanacSolarTerm"

#= 四柱推命用ライブラリ
#== 日干支および月干支から変通星を求める
#== AlmanacOrientalZodiac class と
#== AlmanacSolarTerm class の使用例
#
#Authors::   Yoshihiko Hara(GoroNeko)
#Date::      2010/08/25
#Version::   0.9.0
#Copyright:: Copyright (C) Yoshihiko Hara(GoroNeko), 2010. All rights reserved.
#License::   GNU GENERAL PUBLIC LICENSE Version 2
#

class HentsuStar

  # 本スクリプトの、名前とバージョン
  NAME = "HentsuStar"
  VERSION = "0.9.0"

  # 変通性を示す
  # 変通星番号 → 0(比肩) ～ 9(劫財)
  HENTSU_STAR_NUMBER \
    = Array.new([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

  # 変通星（通変）の漢字和名
  HENTSU_STAR_NAME_J \
    = Array.new(["比肩","正官","傷官","正財","偏財","食神","印綬","偏官","偏印","劫財"])

  # 蔵干テーブル
  # テーブル内の各要素が示す内容は以下の通り
  # 甲->1,乙->2,丙->3,丁->4,戊->5,己->6,庚->7,辛->8,壬->9,癸->0
  # 備考 ： table 構成 → [12][32]
  ZOUKAN_TABLE_SUB0 \
    = Array.new([5,5,5,5,5,5,5,5,1,1,1,1,1,1,1,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9])
  ZOUKAN_TABLE_SUB1 \
    = Array.new([9,9,9,9,9,9,9,9,9,9,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
  ZOUKAN_TABLE_SUB2 \
    = Array.new([0,0,0,0,0,0,0,0,0,0,7,7,7,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6])
  ZOUKAN_TABLE_SUB3 \
    = Array.new([5,5,5,5,5,5,5,5,3,3,3,3,3,3,3,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1])
  ZOUKAN_TABLE_SUB4 \
    = Array.new([1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2])
  ZOUKAN_TABLE_SUB5 \
    = Array.new([2,2,2,2,2,2,2,2,2,2,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5])
  ZOUKAN_TABLE_SUB6 \
    = Array.new([5,5,5,5,5,5,7,7,7,7,7,7,7,7,7,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3])
  ZOUKAN_TABLE_SUB7 \
    = Array.new([3,3,3,3,3,3,3,3,3,3,3,6,6,6,6,6,6,6,6,6,6,4,4,4,4,4,4,4,4,4,4,4])
  ZOUKAN_TABLE_SUB8 \
    = Array.new([4,4,4,4,4,4,4,4,4,4,2,2,2,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6])
  ZOUKAN_TABLE_SUB9 \
    = Array.new([6,6,6,6,6,6,6,6,5,5,5,9,9,9,9,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7])
  ZOUKAN_TABLE_SUB10 \
    = Array.new([7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8])
  ZOUKAN_TABLE_SUB11 \
    = Array.new([8,8,8,8,8,8,8,8,8,8,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5])

  ZOUKAN_TABLE \
    = Array.new([ZOUKAN_TABLE_SUB0, ZOUKAN_TABLE_SUB1, ZOUKAN_TABLE_SUB2, ZOUKAN_TABLE_SUB3,
                 ZOUKAN_TABLE_SUB4, ZOUKAN_TABLE_SUB5, ZOUKAN_TABLE_SUB6, ZOUKAN_TABLE_SUB7,
                 ZOUKAN_TABLE_SUB8, ZOUKAN_TABLE_SUB9, ZOUKAN_TABLE_SUB10, ZOUKAN_TABLE_SUB11])

  # 変通星テーブル
  # 備考 ： table 構成 → [10][10]
  HENTSU_STAR_TABLE_SUB0 = Array.new([0,2,5,3,4,1,7,6,8,9])
  HENTSU_STAR_TABLE_SUB1 = Array.new([6,0,9,5,2,4,3,7,1,8])
  HENTSU_STAR_TABLE_SUB2 = Array.new([8,9,0,2,5,3,4,1,7,6])
  HENTSU_STAR_TABLE_SUB3 = Array.new([1,8,6,0,9,5,2,4,3,7])
  HENTSU_STAR_TABLE_SUB4 = Array.new([7,6,8,9,0,2,5,3,4,1])
  HENTSU_STAR_TABLE_SUB5 = Array.new([3,7,1,8,6,0,9,5,2,4])
  HENTSU_STAR_TABLE_SUB6 = Array.new([4,1,7,6,8,9,0,2,5,3])
  HENTSU_STAR_TABLE_SUB7 = Array.new([2,4,3,7,1,8,6,0,9,5])
  HENTSU_STAR_TABLE_SUB8 = Array.new([5,3,4,1,7,6,8,9,0,2])
  HENTSU_STAR_TABLE_SUB9 = Array.new([9,5,2,4,3,7,1,8,6,0])

  HENTSU_STAR_TABLE \
    = Array.new([HENTSU_STAR_TABLE_SUB0, HENTSU_STAR_TABLE_SUB1, HENTSU_STAR_TABLE_SUB2,
                 HENTSU_STAR_TABLE_SUB3, HENTSU_STAR_TABLE_SUB4, HENTSU_STAR_TABLE_SUB5,
                 HENTSU_STAR_TABLE_SUB6, HENTSU_STAR_TABLE_SUB7, HENTSU_STAR_TABLE_SUB8,
                 HENTSU_STAR_TABLE_SUB9])

  # 指定西暦年( year:Integer )かつ 指定月( month:Integer )かつ
  # 指定日( day:Integer ) の、変通星番号( Integer )を求める
  #
  # 【引数】
  #   1.対象年( year:Integer )
  #   2.対象月( month:Integer )
  #   3.対象日( day:Integer )
  # 【戻り値】
  #   1.変通星番号( Integer )
  #
  def get_hentsu_star_no(year, month, day)

    # 二十四節気算出用の class および 干支算出用の class の
    # 使用準備をする
    solar_term = AlmanacSolarTerm.new
    oriental_zodiac = AlmanacOrientalZodiac.new

    # 直前の節から、対象年月日までの経過日を算出する
    target_date \
      = AlmanacDateTime.new(year,month,day,0,0,0,Rational(3,8))
    day_of_passage_from_setsu \
      = solar_term.get_day_of_passage_from_setsu(target_date)

    # 対象年月日の 月干支 から 月の 十二支番号 を得る
    month_number = oriental_zodiac.get_oriental_zodiac_of_month_number(target_date)
    jyuunishi_number = oriental_zodiac.get_jyuunishi_number(month_number)

    # 対象年月日の 日干支 から 日の 十干番号 を得る
    day_number = oriental_zodiac.get_oriental_zodiac_of_day_number(target_date)
    jikkann_number = oriental_zodiac.get_jikkann_number(day_number)

    # 蔵干番号を求める
    zoukan_number \
      = get_zoukan_number(day_of_passage_from_setsu, jyuunishi_number)

    # 変通星番号を得る
    hentsu_star_number \
      = get_hentsu_star_number(zoukan_number, jikkann_number)

    return hentsu_star_number

  end

  def get_zoukan_number(day_of_passage_from_setsu, jyuunishi_number)

    zoukan_number = ZOUKAN_TABLE[jyuunishi_number][day_of_passage_from_setsu]
    return zoukan_number

  end

  def get_hentsu_star_number(zoukan_number, jikkann_number)

    hentsu_star_no = HENTSU_STAR_TABLE[jikkann_number][zoukan_number]
    return hentsu_star_no

  end

end
